/*
 * CxIO - simple XOR-based encryption mimicking stdio-based file IO
 * Copyright (c) Martin Dietze
 *
 * Unless otherwise stated, this software is provided under the terms of the
 * GNU LESSER PUBLIC LICENSE. The text of this license agreement accompanies
 * this library's source distribution.
 */

#include "cxio.h"
#include <stdio.h>
#include <string.h>
#include <assert.h>

#ifndef __DEBUG
# define __DEBUG 0
#endif

#if __DEBUG > 1
# define TRACE(a) a
# define TRACE2(a) a
#elif __DEBUG > 0
# define TRACE(a) a
# define TRACE2(a)
#else
# define TRACE(a)
# define TRACE2(a)
#endif

struct cx_file
{
  FILE *fptr;
  char key[256];
  int keylen;
  int keyidx;
  int empty;
};

static struct cx_file files[FOPEN_MAX];
static int initialised = 0;

static void p_init (void);
static int p_cx_fputc (int, struct cx_file *);
static int p_cx_fgetc (struct cx_file *);

int
cx_fclose (struct cx_file *cxp)
{
  int rc;
  if (cxp == NULL)
    {
      return fclose (NULL);
    }
  rc = fclose (cxp->fptr);
  cxp->fptr = NULL;
  cxp->key[0] = '\0';
  cxp->keylen = 0;
  cxp->keyidx = 0;
  cxp->empty = 1;
  return rc;
}

struct cx_file *
cx_fopen (const char *name, const char *mode)
{
  int i = 0;
  FILE *fp;
  char my_mode[256];
  char *tpos;

  if (!initialised)
    {
      p_init ();
    }

  while (!files[i].empty && i < FOPEN_MAX)
    {
      ++i;
    }

  if (i == FOPEN_MAX)
    {
      return NULL;
    }

  strcpy (my_mode, mode);
  tpos = strchr (my_mode, 't');

  if (tpos != NULL)
    {
      *tpos = 'b';
    }
  else
    {
      strcat (my_mode, "b");
    }

  if ((fp = fopen (name, my_mode)) == NULL)
    {
      return NULL;
    }

  files[i].empty = 0;
  files[i].fptr = fp;
  files[i].keylen = 0;

  return &files[i];
}

int
cx_feof (struct cx_file *cxp)
{
  return feof (cxp->fptr);
}

int
cx_fputs (const char *str, struct cx_file *cxp)
{
  int i = 0;

  if (str == NULL)
    {
      return EOF;
    }

  while (str[i] != '\0')
    {
      if (p_cx_fputc (str[i], cxp) == EOF)
	{
	  return EOF;
	}
      i++;
    }

  return i;
}

char *
cx_fgets (char *bufp, int count, struct cx_file *cxp)
{
  int i = 0;
  int tmp;

  if (bufp == NULL)
    {
      return bufp;
    }

  while (i < count)
    {
      tmp = p_cx_fgetc (cxp);
      if (cx_feof (cxp))
	{
	  break;
	}
      if (tmp == EOF)
	{
	  return NULL;
	}
      bufp[i] = tmp;
      i++;
      if (bufp[i - 1] == '\n')
	{
	  break;
	}
    }

  if (i != 0)
    {
      bufp[i] = '\0';
    }
  else
    {
      return NULL;
    }

  return bufp;
}

size_t
cx_fwrite (const void *p, size_t size, size_t nmemb, struct cx_file *cxp)
{
  int i = 0;
  int bytes = (int)(size * nmemb);
  const char *bufp = p;

  if (bufp == NULL)
    {
      return 0;
    }

  for (i = 0; i < bytes && p_cx_fputc (bufp[i], cxp) != EOF; ++i)
    {
    }

  TRACE (fprintf (stderr, "cx_fwrite %d\n", i));
  return i;
}

size_t
cx_fread (void *p, size_t size, size_t nmemb, struct cx_file *cxp)
{
  int i = 0;
  int tmp = 0;
  int bytes = (int)(size * nmemb);
  char *bufp = p;

  if (bufp == NULL)
    {
      return 0;
    }

  for (; tmp != EOF && i < bytes; ++i)
    {
      tmp = p_cx_fgetc (cxp);
      if (cx_feof(cxp))
	{
	  tmp = EOF;
	}
      if (tmp != EOF)
	{
	  bufp[i] = tmp;
	}
    }
  if (tmp == EOF)
    {
      --i;
    }

  TRACE (fprintf (stderr, "cx_fread %d, eof: %d\n", i, tmp == EOF));
  return i;
}

int
cx_setkey (const char *key, struct cx_file *cxp)
{
  if (cxp == NULL)
    {
      return -1;
    }

  strncpy (cxp->key, key, 256);
  cxp->key[255] = '\0';
  cxp->keyidx = 0;
  cxp->keylen = strlen (cxp->key);

  return cxp->keylen;
}

static void
p_init (void)
{
  for (; initialised < FOPEN_MAX; initialised++)
    {
      files[initialised].empty = 1;
      files[initialised].key[0] = '\0';
      files[initialised].keyidx = 0;
    }
}

static int
p_cx_fputc (int ch, struct cx_file *cxp)
{
  unsigned char tmp = cxp->keylen > 0
    ? (unsigned char) ch ^ (unsigned char) cxp->key[cxp->keyidx]
    : (unsigned char) ch;
  int rc = fputc (tmp, cxp->fptr);

  TRACE2 (fprintf (stderr, "encoded %c, fputc wrote %d\n", (unsigned char) ch, rc));

  if (rc == EOF)
    {
      return rc;
    }

  if (cxp->keylen > 0)
    {
      cxp->keyidx = (cxp->keyidx + 1) % cxp->keylen;
    }
  return rc;
}

static int
p_cx_fgetc (struct cx_file *cxp)
{
  int rc = fgetc (cxp->fptr);

  TRACE2 (fprintf (stderr, "fgetc read %d, ", rc));

  if (rc == EOF)
    {
      TRACE (fprintf (stderr, "NOT decoded (EOF).\n"));
      return rc;
    }

  if (cxp->keylen > 0)
    {
      rc = (int) ((unsigned char) rc ^ (unsigned char) cxp->key[cxp->keyidx]);
    }

  TRACE2 (fprintf (stderr, "decoded: '%c'\n", (unsigned char) rc));

  if (cxp->keylen > 0)
    {
      cxp->keyidx = (cxp->keyidx + 1) % cxp->keylen;
    }

  assert(rc != EOF);
  return rc;
}
